from __future__ import annotations

from math import cos, isclose, sin, sqrt

import attr
import numpy as np

from transit_chem.config import conf
from transit_chem.elements import Element, element_from_string
from transit_chem.utils import angle


@attr.s(frozen=True, cmp=False)
class Atom:
    """Atom class containing coordinates, basis and mass."""

    element: Element = attr.ib(converter=element_from_string)
    x: float = attr.ib(converter=float)
    y: float = attr.ib(converter=float)
    z: float = attr.ib(converter=float)

    @property
    def mass(self):
        return self.element.z

    @property
    def coords(self):
        return np.array([self.x, self.y, self.z])

    def with_coords(self, x: float, y: float, z: float) -> "Atom":
        return attr.evolve(self, x=x, y=y, z=z)

    def translated(self, x=0.0, y=0.0, z=0.0) -> Atom:
        return attr.evolve(self, x=self.x + x, y=self.y + y, z=self.z + z)

    def scaled(self, factor: float) -> Atom:
        return attr.evolve(self, x=factor * self.x, y=factor * self.y, z=factor * self.z)

    def rotated(self, r: np.ndarray) -> Atom:
        if not r.shape == (3, 3):
            raise ValueError(f"Rotation matrix R must be 3x3, got {r.shape}")
        x, y, z = r @ np.array([self.x, self.y, self.z])
        return attr.evolve(self, x=x, y=y, z=z)

    def flipped_x(self) -> Atom:
        return attr.evolve(self, x=-self.x)

    def distance(self, other: Atom) -> float:
        return sqrt((self.x - other.x) ** 2 + (self.y - other.y) ** 2 + (self.z - other.z) ** 2)

    @property
    def normalized_coords(self) -> np.array:
        return self.coords / np.linalg.norm(self.coords)

    def rotation_matrix_to(self, x: float, y: float, z: float):
        """https://math.stackexchange.com/questions/180418/calculate-rotation-matrix-to-align-vector-a-to-vector-b-in-3d"""
        a = self.normalized_coords
        b = np.array([x, y, z])
        b = b / np.linalg.norm(b)
        v = np.cross(a, b)
        s = np.linalg.norm(v)
        c = np.dot(a, b)
        I = np.eye(*a.shape)  # noqa: E741

        v_x = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        return I + v_x + v_x @ v_x * (1 - c) / s ** 2

    def angle_to_xy_plane(self) -> float:
        return angle(self.coords, [self.x, self.y, 0])

    def angle_to_xz_plane(self) -> float:
        return angle(self.coords, [self.x, 0, self.z])

    def angle_to_yz_plane(self) -> float:
        return angle(self.coords, [0, self.y, self.z])

    def rotated_about_x(self, angle: float) -> "Atom":
        r = np.array([[1, 0, 0], [0, cos(angle), -sin(angle)], [0, sin(angle), cos(angle)]])
        return self.with_coords(*r @ self.coords)

    def rotated_about_z(self, angle: float) -> "Atom":
        r = np.array([[cos(angle), -sin(angle), 0], [sin(angle), cos(angle), 0], [0, 0, 1]])
        return self.with_coords(*r @ self.coords)

    def rotated_about_y(self, angle: float) -> "Atom":
        r = np.array([[cos(angle), 0, -sin(angle)], [0, 1, 0], [sin(angle), 0, cos(angle)]])
        return self.with_coords(*r @ self.coords)

    def __eq__(self, other: "Atom") -> bool:
        return (
            self.element is other.element
            and isclose(self.x, other.x, abs_tol=conf.small_number)
            and isclose(self.y, other.y, abs_tol=conf.small_number)
            and isclose(self.z, other.z, abs_tol=conf.small_number)
        )



from __future__ import annotations

from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from itertools import combinations_with_replacement, product
from math import acos
from math import isclose as math_isclose
from math import sqrt
from typing import Callable, Dict, Iterable, Sequence, Tuple, TypeVar

import attr
import numpy as np
from joblib import Memory
from tqdm import tqdm

from transit_chem.config import conf
from transit_chem.validation import Range, not_inf, not_nan

__all__ = ["pairwise_array_from_func", "Parabola", "LinearComb", "cache"]


memory = Memory(conf.cache_dir, verbose=conf.joblib_verbosity)
cache = memory.cache

T = TypeVar("T")


def isclose(x: float, y: float) -> bool:
    return math_isclose(x, y, abs_tol=conf.small_number)


class LinearComb:
    def __init__(self, c, f):
        self.c = c
        self.f = f

    def __call__(self, x):
        return sum(ci * fi(x) for ci, fi in zip(self.c, self.f))

    def __repr__(self):
        return " + ".join(f"{c} * {f}" for c, f in zip(self.c, self.f))

    def __kinetic__(self):
        return LinearComb(c=self.c, f=[f.__kinetic__() for f in self.f])


@cache
def pairwise_array_from_func(
    items: Sequence[T], func: Callable[[T, T], float], symmetric=False
) -> np.ndarray:
    """Create a pairwise array by applying a function to all pairs of items.


    Parameters
    -----------
    items
        A container from which pairs will be generated. Must support len() and integer indexing over range(len(items))
    func
        A function f(first, second, *args, **kwargs) which takes 2 items and returns a float.
    symmetric
        Whether the resulting matrix should be symmetric. If true,
        will only compute each (i, j) pair once and set both [i, j] and [j, i] to that value.

    Returns
    --------
    np.array
        The resulting matrix

    Examples
    ---------

    >>> from transit_chem.utils import pairwise_array_from_func
    >>> def distance(i, j):
    ...     return abs(i - j)
    ...
    >>> pairwise_array_from_func([1, 2, 4], distance)
    array([[0., 1., 3.],
           [1., 0., 2.],
           [3., 2., 0.]])
    >>> pairwise_array_from_func([1, 2, 4], distance, symmetric=True)
    array([[0., 1., 3.],
           [1., 0., 2.],
           [3., 2., 0.]])


    """
    n = len(items)
    result = np.zeros((n, n))

    fut_res: Dict[Future, Tuple] = {}

    if symmetric:
        combs = list(combinations_with_replacement(range(n), 2))
    else:
        combs = list(product(range(n), repeat=2))

    with ProcessPoolExecutor() as exc:
        for i, j in combs:
            fut_res[exc.submit(func, items[i], items[j])] = (i, j)

        with tqdm(total=len(combs), disable=not conf.enable_progressbar, ncols=100) as pbar:
            for future in as_completed(fut_res.keys()):
                res = future.result()
                i, j = fut_res[future]
                if symmetric:
                    result[i, j] = result[j, i] = res
                else:
                    result[i, j] = res
                pbar.update(1)

    return result


@attr.s(frozen=True)
class Parabola:
    a: float = attr.ib(validator=[not_nan, not_inf, Range(-conf.large_number, conf.large_number)])
    b: float = attr.ib(validator=[not_nan, not_inf, Range(-conf.large_number, conf.large_number)])
    c: float = attr.ib(validator=[not_nan, not_inf, Range(-conf.large_number, conf.large_number)])

    def __call__(self, x):
        return self.a * x ** 2 + self.b * x + self.c

    @property
    def has_vertex(self) -> bool:
        return not isclose(self.a, 0.0)

    @property
    def vertex(self) -> float:
        if not self.has_vertex:
            raise ValueError(f"{self} is a line and has no vertex.")
        return -self.b / (2.0 * self.a)

    @staticmethod
    def from_points(
        point1: Tuple[float, float], point2: Tuple[float, float], point3: Tuple[float, float]
    ) -> Parabola:
        """Create a Parabola passing through 3 points.

        Parameters
        -----------
        point1
            x, y point
        point2
            x, y point
        point3
            x, y point

        Returns
        --------
        Parabola
            Parabola with coefficients fit to the points.

        """
        x1, y1 = point1
        x2, y2 = point2
        x3, y3 = point3

        # The polynomial coefficients are the solution to the
        # linear equation Ax = b where
        # x = [a, b, c]
        A = np.array([[x1 ** 2, x1, 1], [x2 ** 2, x2, 1], [x3 ** 2, x3, 1]])
        b = np.array([y1, y2, y3])
        coeffs = np.linalg.solve(A, b)

        return Parabola(a=coeffs[0], b=coeffs[1], c=coeffs[2])


def dot(v1: Iterable[float], v2: Iterable[float]) -> float:
    return sum((a * b) for a, b in zip(v1, v2))


def length(v: Iterable[float]) -> float:
    return sqrt(dot(v, v))


def angle(v1: Iterable[float], v2: Iterable[float]) -> float:
    return acos(dot(v1, v2) / (length(v1) * length(v2)))



from __future__ import annotations

import math
from itertools import count, takewhile
from typing import Callable, List, Tuple

import attr

# import numba
import numpy as np
from numpy.linalg import inv
from scipy import special
from scipy.linalg import eigh

from transit_chem.config import conf
from transit_chem.operators import Overlap
from transit_chem.potentials import Harmonic
from transit_chem.utils import LinearComb, Parabola, cache, isclose
from transit_chem.validation import Range

___all__ = ["HarmonicOscillatorWaveFunction", "overlap1d"]


# class ThreeDBasis:
#     def __init__(self, factory, x0, y0, z0, alpha):
#         """Create a basis object from some parameters and a function-factory function.
#
#         params
#         ======
#             factory: function g(x0, y0, z0, alpha) that returns a function f(x, y, z)
#             x0, y0, z0: centers of basis function
#             alpha: parameter for basis (decay coefficient or some other such nonsense).
#         """
#         self.f = factory(x0, y0, z0, alpha)
#         self.x0 = x0
#         self.y0 = y0
#         self.z0 = z0
#         self.alpha = alpha
#
#     def __call__(self, x, y, z):
#         return self.f(x, y, z)
#
#     def overlap(
#         self,
#         other,
#         distance_cutoff=5.0,
#         interval=5.0,
#         xmin=None,
#         xmax=None,
#         ymin=None,
#         ymax=None,
#         zmin=None,
#         zmax=None,
#     ):
#         """Return the overlap of this with another
#
#         This is to reduce the calculations performed by naively integrating over
#         all space for all combinations of basis functions.
#         Return 0.0 if the basis functions are separated by more than distance_cutoff.
#         Integrate from interval away from the centers. If bounds are not given.
#
#         If the midpoint of the centers of the functions is more than distance_cutoff from the boundaries,
#         0.0 is returned.
#
#         """
#         if not isinstance(other, ThreeDBasis):
#             raise TypeError(f"ThreeDBasis can't be overlapped with {type(other)}")
#         r = np.sqrt(
#             (self.x0 - other.x0) ** 2 + (self.y0 - other.y0) ** 2 + (self.z0 - other.z0) ** 2
#         )
#         if r > distance_cutoff:
#             return 0.0
#         else:
#             if xmin is None:
#                 xmin = min(self.x0, other.x0) - interval
#             if ymin is None:
#                 ymin = min(self.y0, other.y0) - interval
#             if zmin is None:
#                 zmin = min(self.z0, other.z0) - interval
#             if xmax is None:
#                 xmax = max(self.x0, other.x0) + interval
#             if ymax is None:
#                 ymax = max(self.y0, other.y0) + interval
#             if zmax is None:
#                 zmax = max(self.z0, other.z0) + interval
#
#             #  Check to see if the midpoint of the overlapped functions
#             # falls outside integration range.
#             #  The maximum value between the centers,
#             # and if it's sufficiently far from the region of integration
#             #  we can return 0.0 and not lose anything.
#
#             midpoint = (
#                 0.5 * (self.x0 + other.x0),
#                 0.5 * (self.y0 + other.y0),
#                 0.5 * (self.z0 + other.z0),
#             )
#             if midpoint[0] > (xmax + interval) or midpoint[0] < (xmin - interval):
#                 return 0.0
#             if midpoint[1] > (ymax + interval) or midpoint[1] < (ymin - interval):
#                 return 0.0
#             if midpoint[2] > (zmax + interval) or midpoint[2] < (zmin - interval):
#                 return 0.0
#
#             # Finally actually do the overlap if none of the zero-conditions apply
#             ovlp = overlap_factory(self.f, other.f)
#             integrator = integrate_region_gen_3D([[xmin, xmax], [ymin, ymax], [zmin, zmax]])
#             return integrator(ovlp)
#
#
# def overlap_factory(f1, f2):
#     """Generate a jitted function of f1(x, y, z) * f2(x, y, z). Only works on f1, f2 that are numba jitted."""
#
#     @numba.jit("float64(float64, float64, float64)", nopython=True)
#     def overlap_function(x, y, z):
#         return f1(x, y, z) * f2(x, y, z)
#
#     return overlap_function
#
#
# @attr.s()
# class PzOrbital:
#     x: float = attr.ib()
#     y: float = attr.ib()
#     z: float = attr.ib()
#     alpha: float = attr.ib()
#
#     @property
#     def function(self) -> Callable[[float, float, float], float]:
#         """Return a function for pz orbital function centered at (x0, y0, z0) and decay coeff alpha.
#
#         The pz orbital is compiled at runtime using numba's jit LLVM compiler. As such, it is fast!
#         """
#         N = self.alpha ** 2.5 / math.sqrt(math.pi)
#
#         @numba.jit("float64(float64, float64, float64)", nopython=True)
#         def pz_orbital(x, y, z):
#             r = math.sqrt((x - self.x) ** 2 + (y - self.y) ** 2 + (z - self.z) ** 2)
#             return N * (z - self.z) * math.exp(-self.alpha * r)
#
#         return pz_orbital


@attr.s(frozen=True)
class HarmonicOscillator:
    """A 1D quantum harmonic oscillator wave function.

    Parameters
    -----------
    n
        The quantum number
    center
        The center of the potential well
    omega
        The angular frequency of the oscillator
    mass
        The mass of the particle

    Examples
    ---------

    >>> ho = HarmonicOscillator(n=2, center=0.5)
    >>> ho
    HarmonicOscillator(n=2, center=0.5, omega=1, mass=1.0)
    >>> ho(0.5)
    -0.5311259660135984
    >>> ho(1000)
    0.0
    >>> ho(-1000)
    0.0

    """

    n: int = attr.ib(validator=[Range(0, conf.harmonic_oscillator_max_n)])
    center: float = attr.ib(validator=[Range(-conf.large_number, conf.large_number)])
    mass: float = attr.ib(default=1.0, validator=Range(conf.small_number, conf.large_number))
    omega: float = attr.ib(default=1.0, validator=Range(conf.small_number, conf.large_number))

    @staticmethod
    def from_parabola(p: Parabola, n: int, mass: float = 1.0) -> HarmonicOscillator:
        # a = m/2 * w**2
        # 2a / m = w**2
        # sqrt(2a/m) = w
        w = np.sqrt(2 * p.a / mass)
        return HarmonicOscillator(n=n, center=p.vertex, omega=w, mass=mass)

    @staticmethod
    def from_potential_points(
        point1: Tuple[float, float],
        point2: Tuple[float, float],
        point3: Tuple[float, float],
        n: int,
        mass: float = 1.0,
    ) -> HarmonicOscillator:
        """Create a harmonic oscillator wave function from 3 samples of the potential.

        The three points are fit to a parabola (harmonic potential), then the parameters
        for the harmonic oscillator are determined, and the corresponding wave function
        generated and returned.

        Parameters
        -----------
        point1
            A sample point of the potential
        point2
            A sample point of the potential
        point3
            A sample point of the potential
        n
            The quantum number of the resulting wave function
        mass
            The mass of the particle

        Returns
        --------
        HarmonicOscillator
            The harmonic oscillator wave function resulting from the potential at the 3
            points.

        Examples
        ---------

        >>> ho = HarmonicOscillator.from_potential_points(
        ...     point1=(0.5, 1),
        ...     point2=(2.0, 0.5),
        ...     point3=(3.0, 1.5),
        ...     n=0
        ... )
        >>> ho
        HarmonicOscillator(n=0, center=1.5624999999999998, omega=1.0327955589886444, mass=1.0)


        """
        p = Parabola.from_points(point1, point2, point3)
        return HarmonicOscillator.from_parabola(p, n=n, mass=mass)

    def __kinetic__(self):
        """Return kinetic energy operator applied on this.

        K = p^2 / 2m
        p = i * sqrt(m w hbar/2)(a+ - a)

        k = -1/2 * (m w hbar / 2)[(a+ - a)^2]
        [] = (a+^2 + a+a + aa+  - a^2)

        [] = sqrt(n+1)sqrt(n+2)| n + 2 >  always
                sqrt(n+1)sqrt(n+1)| n >      always
                sqrt(n)  sqrt(n)  | n >      if n == 0  0
                sqrt(n)  sqrt(n-1)| n - 2 >  if n <= 1  0

        k = - (m w hbar) / 4 * []
        """

        def k(x):
            first = np.sqrt(self.n + 1) * np.sqrt(self.n + 2) * attr.evolve(self, n=self.n + 2)(x)

            if self.n == 0:
                # Lowering operator on n=0
                # means the second term is zero
                inner = (self.n + 1) * self(x)
            else:
                inner = (2 * self.n + 1) * self(x)

            if self.n <= 1:
                # Lowering operator twice on n=0 or n=1 will be 0
                # if x is numpy array, we want to return a numpy
                # array of zeros, so multiply x by zeros will work on
                # numpy array or float.
                last = 0.0 * x
            else:
                last = np.sqrt(self.n) * np.sqrt(self.n - 1) * attr.evolve(self, n=self.n - 2)(x)

            return -self.mass * self.omega * (first - inner + last) / 4.0

        return k

    def __overlap__(self, other, lower_limit: float, upper_limit: float) -> float:
        if lower_limit != -np.inf or upper_limit != np.inf:
            raise NotImplementedError
        if not isinstance(other, HarmonicOscillator):
            raise NotImplementedError

        if (
            isclose(self.center, other.center)
            and isclose(self.mass, other.mass)
            and isclose(self.omega, other.omega)
        ):
            return 1.0 if self.n == other.n else 0.0
        raise NotImplementedError

    @property
    def energy(self):
        return (self.n + 0.5) * self.omega

    @property
    def potential(self):
        return Harmonic(center=self.center, mass=self.mass, omega=self.omega)

    @property
    def N(self):
        return (
            1.0
            / math.sqrt(2 ** self.n * math.factorial(self.n))
            * ((self.mass * self.omega) / math.pi) ** 0.25
        )

    @property
    def _hermite(self):
        return special.hermite(self.n)

    def __call__(self, x):
        y = (np.sqrt(self.mass * self.omega)) * (x - self.center)
        return self.N * np.exp(-(y ** 2) / 2.0) * self._hermite(y)


def harmonic_basis_from_parabola(p: Parabola, cutoff_energy: float) -> List[HarmonicOscillator]:
    hos = (HarmonicOscillator.from_parabola(p, n=i) for i in count())
    return list(takewhile(lambda ho: ho.energy <= cutoff_energy, hos))


@attr.s(frozen=True)
class EigenBasis:
    # TODO: validate shapes, properties
    states: Tuple[Callable, ...] = attr.ib(converter=tuple)
    energies: Tuple[float, ...] = attr.ib(converter=tuple)
    ao_S: np.array = attr.ib()
    eigvecs: np.array = attr.ib()
    ao_basis: List[Callable] = attr.ib()

    def __len__(self):
        return len(self.states)

    @staticmethod
    def from_basis(basis: List[Callable], H: np.array, S: np.array) -> EigenBasis:
        # Sort vals/vecs
        eigvals, eigvecs = eigh(H, S)
        idx = np.argsort(eigvals)
        eigvals = eigvals[idx]
        eigvecs = eigvecs[:, idx]

        states = [LinearComb(c, basis) for c in eigvecs.T]
        return EigenBasis(
            states=tuple(states), energies=tuple(eigvals), eigvecs=eigvecs, ao_S=S, ao_basis=basis
        )

    def transformed(self, matrix: np.array) -> np.array:
        return inv(self.eigvecs) @ inv(self.ao_S) @ matrix @ self.eigvecs


@cache
def get_expansion_coeffs(state: Callable, basis: List[Callable]) -> List[float]:
    return [Overlap()(state, basis_func) for basis_func in basis]



from typing import Callable, Union

import attr
import numpy as np

from transit_chem.basis import EigenBasis, get_expansion_coeffs
from transit_chem.operators import Operator


@attr.s
class TimeEvolvingState:
    initial_state: Callable = attr.ib()
    eigen_basis: EigenBasis = attr.ib(repr=False)

    def __attrs_post_init__(self):
        self.expansion_coeffs = get_expansion_coeffs(self.initial_state, self.eigen_basis.states)

    def __call__(self, x: Union[float, np.ndarray], t: float) -> float:
        return sum(
            [
                c * np.exp(-1j * e * t) * state(x)
                for state, e, c in zip(
                    self.eigen_basis.states, self.eigen_basis.energies, self.expansion_coeffs
                )
            ]
        )

    def observable(self, operator: Callable, hermitian: bool) -> "TimeEvolvingObservable":
        return TimeEvolvingObservable(self, operator, hermitian=hermitian)


@attr.s()
class TimeEvolvingObservable:
    time_evolving_state: TimeEvolvingState = attr.ib()
    operator: Operator = attr.ib()
    hermitian: bool = attr.ib()

    def __attrs_post_init__(self):
        e = self.time_evolving_state.eigen_basis.energies
        N = len(self.time_evolving_state.eigen_basis)
        c = self.time_evolving_state.expansion_coeffs

        # TODO: this is awkward
        ao_matrix = self.operator.matrix(self.time_evolving_state.eigen_basis.ao_basis)
        eigen_basis_matrix = self.time_evolving_state.eigen_basis.transformed(ao_matrix)

        # The following code is equivalent to the definition of f below.
        # f is written that way to optimize for speed.
        # On my machine, it was roughly 7x faster
        #
        # def f(t: float):
        #     return sum(
        #         c[i] * c[j] * np.exp(1j * (e[j] - e[i]) * t) * eigen_basis_matrix[i, j]
        #         for i in range(N)
        #         for j in range(N)
        #     )

        # Precalculate reused terms
        # P is a real matrix, but we cast it to complex here so that it doesn't
        # get casted inside the function during the multiplication.
        self.P = np.zeros_like(eigen_basis_matrix, dtype=np.complex128)
        self.W = np.zeros_like(eigen_basis_matrix, dtype=np.complex128)
        for i in range(N):
            for j in range(N):
                self.P[i, j] = c[i] * c[j] * eigen_basis_matrix[i, j]
                self.W[i, j] = np.exp(1j * (e[j] - e[i]))

    def __call__(self, t: float) -> float:
        return np.abs(np.sum(self.P * (self.W ** t)))



from typing import Optional, Union

import attr


@attr.s(frozen=True)
class Element:
    z: int = attr.ib(repr=False)
    name: str = attr.ib()
    voie: Optional[float] = attr.ib(default=None, repr=False)
    alpha: Optional[float] = attr.ib(default=None, repr=False)


H = Element(z=1, name="H")
He = Element(z=2, name="He")
Li = Element(z=3, name="Li")
Be = Element(z=4, name="Be")
B = Element(z=5, name="B")
C = Element(z=6, name="C", voie=10.77, alpha=1.5679)
N = Element(z=7, name="N", voie=13.9, alpha=1.9170)
O = Element(z=8, name="O", voie=15.8, alpha=2.2266)  # noqa: E741
F = Element(z=9, name="F")
Ne = Element(z=10, name="Ne")
Na = Element(z=11, name="Na")
Mg = Element(z=12, name="Mg")
Al = Element(z=13, name="Al")
Si = Element(z=14, name="Si")
P = Element(z=15, name="P")
S = Element(z=16, name="S")
Cl = Element(z=17, name="Cl")
Ar = Element(z=18, name="Ar")
K = Element(z=19, name="K")
Ca = Element(z=20, name="Ca")
Sc = Element(z=21, name="Sc")
Ti = Element(z=22, name="Ti")
V = Element(z=23, name="V")
Cr = Element(z=24, name="Cr")
Mn = Element(z=25, name="Mn")
Fe = Element(z=26, name="Fe")
Co = Element(z=27, name="Co")
Ni = Element(z=28, name="Ni")
Cu = Element(z=29, name="Cu")
Zn = Element(z=30, name="Zn")
Ga = Element(z=31, name="Ga")
Ge = Element(z=32, name="Ge")
As = Element(z=33, name="As")
Se = Element(z=34, name="Se")
Br = Element(z=35, name="Br")
Kr = Element(z=36, name="Kr")
Rb = Element(z=37, name="Rb")
Sr = Element(z=38, name="Sr")
Y = Element(z=39, name="Y")
Zr = Element(z=40, name="Zr")
Nb = Element(z=41, name="Nb")
Mo = Element(z=42, name="Mo")
Tc = Element(z=43, name="Tc")
Ru = Element(z=44, name="Ru")
Rh = Element(z=45, name="Rh")
Pd = Element(z=46, name="Pd")
Ag = Element(z=47, name="Ag")
Cd = Element(z=48, name="Cd")
In = Element(z=49, name="In")
Sn = Element(z=50, name="Sn")
Sb = Element(z=51, name="Sb")
Te = Element(z=52, name="Te")
I = Element(z=53, name="I")  # noqa: E741
Xe = Element(z=54, name="Xe")
Cs = Element(z=55, name="Cs")
Ba = Element(z=56, name="Ba")
La = Element(z=57, name="La")
Ce = Element(z=58, name="Ce")
Pr = Element(z=59, name="Pr")
Nd = Element(z=60, name="Nd")
Pm = Element(z=61, name="Pm")
Sm = Element(z=62, name="Sm")
Eu = Element(z=63, name="Eu")
Gd = Element(z=64, name="Gd")
Tb = Element(z=65, name="Tb")
Dy = Element(z=66, name="Dy")
Ho = Element(z=67, name="Ho")
Er = Element(z=68, name="Er")
Tm = Element(z=69, name="Tm")
Yb = Element(z=70, name="Yb")
Lu = Element(z=71, name="Lu")
Hf = Element(z=72, name="Hf")
Ta = Element(z=73, name="Ta")
W = Element(z=74, name="W")
Re = Element(z=75, name="Re")
Os = Element(z=76, name="Os")
Ir = Element(z=77, name="Ir")
Pt = Element(z=78, name="Pt")
Au = Element(z=79, name="Au")
Hg = Element(z=80, name="Hg")
Tl = Element(z=81, name="Tl")
Pb = Element(z=82, name="Pb")
Bi = Element(z=83, name="Bi")
Po = Element(z=84, name="Po")
At = Element(z=85, name="At")
Rn = Element(z=86, name="Rn")
Fr = Element(z=87, name="Fr")
Ra = Element(z=88, name="Ra")
Ac = Element(z=89, name="Ac")
Th = Element(z=90, name="Th")
Pa = Element(z=91, name="Pa")
U = Element(z=92, name="U")
Np = Element(z=93, name="Np")
Pu = Element(z=94, name="Pu")
Am = Element(z=95, name="Am")
Cm = Element(z=96, name="Cm")
Bk = Element(z=97, name="Bk")
Cf = Element(z=98, name="Cf")
Es = Element(z=99, name="Es")
Fm = Element(z=100, name="Fm")
Md = Element(z=101, name="Md")
No = Element(z=102, name="No")
Lr = Element(z=103, name="Lr")
Rf = Element(z=104, name="Rf")
Db = Element(z=105, name="Db")
Sg = Element(z=106, name="Sg")
Bh = Element(z=107, name="Bh")
Hs = Element(z=108, name="Hs")
Mt = Element(z=109, name="Mt")
Ds = Element(z=110, name="Ds")
Rg = Element(z=111, name="Rg")
Cn = Element(z=112, name="Cn")
Uut = Element(z=113, name="Uut")
Fl = Element(z=114, name="Fl")
Uup = Element(z=115, name="Uup")
Lv = Element(z=116, name="Lv")
Uus = Element(z=117, name="Uus")
Uuo = Element(z=118, name="Uuo")


all_elements = [e for e in globals().values() if isinstance(e, Element)]


def element_from_string(s: Union[str, int, Element]) -> Element:
    if isinstance(s, Element):
        return s

    try:
        z = int(s)
        return next(e for e in all_elements if e.z == z)
    except ValueError:
        return next(e for e in all_elements if e.name.lower() == s.lower())



from pathlib import Path
from typing import Callable, Iterable, List, Union

import numpy as np
from matplotlib import pyplot as plt


def plot_occupancies_over_time(
    times: Iterable[float], occ_probs: Iterable[Callable], save_to: Union[str, Path, None] = None,
):
    fig = plt.figure()
    t = np.array(times)

    s_list = [np.array([s(t) for t in times]) for s in occ_probs]

    for i, s in enumerate(s_list):
        plt.plot(t, s, label=f"Well {i+1}")
    plt.plot(t, sum(s_list), "k--", label="Sum")
    plt.legend()
    if save_to is not None:
        plt.savefig(save_to)
    return fig


def plot_p_not(
    p_not: Callable,
    times: List[float],
    acceptor_occ_prob: Callable,
    save_to: Union[str, Path, None] = None,
):
    fig = plt.figure()

    occprob = np.array([1.0 - acceptor_occ_prob(t) for t in times])
    pnot_over_time = [p_not(t) for t in times]

    plt.plot(times, pnot_over_time, label="pnot")
    plt.plot(times, occprob, label="occprob")
    plt.legend()
    if save_to is not None:
        plt.savefig(save_to)
    return fig


def plot_matrix(
    f: Callable[[np.array], np.array],
    x: List[float],
    save_to: Union[str, Path, None] = None,
    axis: int = 0,
):
    allowed = (0, 1, 2)
    if axis not in allowed:
        raise ValueError(f"Axis must be in {allowed}, got {axis}")

    mat = f(np.array(x))
    if not len(mat.shape) == 3:
        raise ValueError(f"Array must be 3d")

    M, N = [x for i, x in enumerate(mat.shape) if i != axis]
    fig, axes = plt.subplots(M, N)

    for i in range(M):
        for j in range(N):
            if axis == 0:
                axes[i, j].plot(x, mat[:, i, j], label=f"{i}{j}")
            elif axis == 1:
                axes[i, j].plot(x, mat[i, :, j], label=f"{i}{j}")
            elif axis == 2:
                axes[i, j].plot(x, mat[i, j, :], label=f"{i}{j}")

    plt.legend()
    if save_to is not None:
        plt.savefig(save_to)
    return fig, axes



"""Console script for transit_chem."""
import sys
from functools import partial
from pathlib import Path

import click
import numpy as np
from loguru import logger

from transit_chem import config
from transit_chem.basis import EigenBasis, HarmonicOscillator, harmonic_basis_from_parabola
from transit_chem.hopping_matrix import HoppingMatrix, OccupancyProbabilites, Pnot
from transit_chem.operators import Hamiltonian, Overlap
from transit_chem.plotting import plot_matrix, plot_occupancies_over_time, plot_p_not
from transit_chem.potentials import TripleWell
from transit_chem.time_evolution import TimeEvolvingState


@click.group()
@click.option(
    "-l",
    "--log-level",
    type=click.Choice(["TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"]),
    default="INFO",
)
def cli(log_level: str):
    logger.remove()
    logger.add(sys.stdout, level=log_level)


@cli.command("run-oned")
@click.option("-o", "--output-path", type=str, default=".", show_default=True)
@click.option(
    "--cutoff-energy",
    type=float,
    help="The cutoff energy for the basis set. Basis functions will all be below this energy",
    default=4.0,
    show_default=True,
)
@click.option(
    "--until-pnot",
    type=float,
    help="The cutoff value of Pnot to calculate to. For tau90, use --until-pnot=0.1",
    default=0.9,
    show_default=True,
)
@click.option(
    "--pnot_conv_tol",
    type=float,
    default=1.0,
    help="Tolerance for converging until-pnot value with respect to time step.",
    show_default=True,
)
@click.option(
    "--acceptor",
    type=click.Choice(["1", "2", "3"]),
    default="3",
    help="Which well is the acceptor. ",
    show_default=True,
)
@click.option("--well1_depth", type=float, default=0.7, show_default=True)
@click.option("--well1_halfwidth", type=float, default=1.0, show_default=True)
@click.option("--bridge_length", type=float, default=8.0, show_default=True)
@click.option("--bridge_depth", type=float, default=0.5, show_default=True)
@click.option("--well3_halfwidth", type=float, default=1.0, show_default=True)
@click.option("--well3_depth", type=float, default=0.5, show_default=True)
def oned(
    output_path: str = ".",
    log_level: str = "INFO",
    cutoff_energy: float = 4.0,
    until_pnot: float = 0.9,
    pnot_conv_tol: float = 1.0,
    acceptor: str = 3,
    well1_depth=0.7,
    well1_halfwidth=1.0,
    bridge_length=8.0,
    bridge_depth=0.5,
    well3_halfwidth=1.0,
    well3_depth=0.5,
):
    logger.info("Running transit time calculation")

    triple_well = TripleWell.from_params(
        well1_depth=well1_depth,
        well1_halfwidth=well1_halfwidth,
        bridge_length=bridge_length,
        bridge_depth=bridge_depth,
        well3_halfwidth=well3_halfwidth,
        well3_depth=well3_depth,
    )

    well1_basis = harmonic_basis_from_parabola(triple_well.well1, cutoff_energy=cutoff_energy)
    well2_basis = harmonic_basis_from_parabola(triple_well.well2, cutoff_energy=cutoff_energy)
    well3_basis = harmonic_basis_from_parabola(triple_well.well3, cutoff_energy=cutoff_energy)
    basis = well1_basis + well2_basis + well3_basis
    logger.info(
        f"Basis set size is well1: {len(well1_basis)} well2: {len(well2_basis)}"
        f" well3: {len(well3_basis)} a total size of {len(basis)}"
    )

    logger.info("Calculating the Hamiltonian")
    H = Hamiltonian(triple_well).matrix(basis)

    logger.info("Calculating the overlap matrix")
    S = Overlap().matrix(basis)

    logger.info("Calculating the eigen basis")
    eig_basis = EigenBasis.from_basis(basis, H, S)

    logger.info("Time evolving the initial state")
    time_evolving = TimeEvolvingState(
        eigen_basis=eig_basis, initial_state=HarmonicOscillator(n=0, center=triple_well.center1[0])
    )
    occ_probs = OccupancyProbabilites.from_1d_state(
        time_evolving, borders=(-np.inf, triple_well.barrier12[0], triple_well.barrier23[0], np.inf)
    )

    hopping_matrix = HoppingMatrix(occ_probs)

    logger.info("Starting p_not generator computation")
    p_not = Pnot.converged_with_timestep(
        hopping_matrix, until_equals=until_pnot, tolerance=pnot_conv_tol, acceptor=int(acceptor) - 1
    )

    output_path = Path(output_path)
    plot_occupancies_over_time(
        np.array(p_not.times), occ_probs, save_to=output_path / "occupancies_over_time.png"
    )
    plot_p_not(p_not, p_not.times, p_not.acceptor_occ_prob, save_to=output_path / "p_not.png")

    plot_matrix(
        partial(hopping_matrix, delta_t=p_not.delta_t),
        p_not.times,
        save_to=output_path / "hopping_matrix.png",
    )


@cli.group("config")
def config_():
    pass


config_.command("show")(config.show)
config_.command("restore-defaults")(config.restore_defaults)


@config_.command("set")
@click.argument("args", nargs=-1)
def set_items(args):
    kwargs = dict(s.split("=") for s in args)
    config.set_items(**kwargs)


@cli.command("clear-cache")
def clear_cache():
    click.confirm(
        f"This will remove the contents of {config.conf.cache_dir}, recursively. Continue?",
        abort=True,
    )
    config.clear_cache()


if __name__ == "__main__":
    cli()



from functools import lru_cache
from itertools import chain, count, islice, takewhile, tee
from math import isclose
from typing import Callable, Generator, Iterable, Iterator, List, Tuple, Union

import attr
import numpy as np
from loguru import logger
from scipy.interpolate import interp1d
from scipy.optimize import root_scalar

from transit_chem.operators import Overlap
from transit_chem.time_evolution import TimeEvolvingObservable, TimeEvolvingState


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


@attr.s(frozen=True)
class OccupancyProbabilites:
    s: Tuple[TimeEvolvingObservable] = attr.ib(converter=tuple)

    @property
    def initial(self) -> np.array:
        return self.__call__(0)

    def __call__(self, t: float) -> np.array:
        return np.array([s(t) for s in self.s])

    def __getitem__(self, item) -> Callable:
        return self.s[item]

    def __iter__(self):
        return iter(self.s)

    @staticmethod
    def from_1d_state(
        state: TimeEvolvingState, borders: Tuple[float, ...]
    ) -> "OccupancyProbabilites":

        bounds = pairwise(borders)
        overlaps = [Overlap(a, b) for a, b in bounds]
        return OccupancyProbabilites(
            tuple(state.observable(ovlp, hermitian=True) for ovlp in overlaps)
        )


@attr.s(frozen=True)
class HoppingMatrix:
    occ_probs: OccupancyProbabilites = attr.ib()
    N: int = 3

    def at_time(self, t: float, delta_t: float) -> np.array:
        d_occ_probs = self.occ_probs(t) - self.occ_probs(t - delta_t)
        increasing = d_occ_probs > 0
        two_states_increasing = np.count_nonzero(increasing) == 2

        A = np.zeros((self.N, self.N))
        if two_states_increasing:
            # If the state is decreasing in probability, the diagonal is
            # determined
            for j in range(3):
                if not increasing[j]:
                    A[j, j] = self.occ_probs[j](t) / self.occ_probs[j](t - delta_t)
                    for k in range(3):
                        if k != j:
                            A[k, j] = d_occ_probs[k] / self.occ_probs[j](t)

                # If the state is increasing, diagonal is 1. Off diagonal
                # columns = 0
                elif increasing[j] > 0:
                    A[j, j] = 1
                    for k in range(3):
                        if k != j:
                            A[k, j] = 0

        else:
            for j in range(3):
                # If the state is decreasing in probability, the diagonal
                # is determined
                if not increasing[j]:
                    A[j, j] = self.occ_probs[j](t) / self.occ_probs[j](t - delta_t)
                    for k in range(3):
                        if k != j:
                            A[j, k] = 0

            # If the state is increasing, diagonal is 1. Off diagonal
            # column element is 1 - diagonal
            for j in range(3):
                if increasing[j]:
                    A[j, j] = 1
                    for k in range(3):
                        if k != j:
                            A[j, k] = 1 - A[k, k]

        return A

    def __call__(self, t: Union[float, np.ndarray], delta_t: float):
        if isinstance(t, Iterable):
            result = [self.at_time(t_, delta_t) for t_ in t]
            if len(result) == 0:
                raise ValueError("Call to hopping matrix had no elements in the iterable.")
            return np.stack(result)
        else:
            return self.at_time(t, delta_t)


def accumulate(iterable, func, *, initial=None):
    """https://docs.python.org/3/library/itertools.html#itertools.accumulate

    Stolen but not used since I ddidn have ython 3.8 handy and wanted to use the initial
    keyword argument

    """
    it = iter(iterable)
    total = initial
    if initial is None:
        try:
            total = next(it)
        except StopIteration:
            return
    yield total
    for element in it:
        total = func(total, element)
        yield total


@attr.s(frozen=True)
class Pnot:
    acceptor: int = attr.ib()
    hopping_matrix: HoppingMatrix = attr.ib(hash=False)
    times: List[float] = attr.ib()
    values: List[float] = attr.ib()
    delta_t: float = attr.ib()

    """Determine P_not, the probability than acceptor state has never been occupied.

    Parameters
    -----------
    acceptor
        The index of the acceptor well
    hopping_matrix
        A (N, 3, 3) array containing the hopping matrix over time
    p0
        An (3,) array containing probabilities of occupying each region initially

    """

    @property
    def min_time(self) -> float:
        return self.times[0]

    @property
    def max_time(self) -> float:
        return self.times[-1]

    @property
    def tau90(self) -> float:
        return self.time_when_equal_to(0.1)

    @property
    def initial_value(self) -> float:
        return self(self.min_time)

    @property
    def final_value(self) -> float:
        return self(self.max_time)

    @lru_cache()
    def time_when_equal_to(self, x: float) -> float:
        def f(t):
            return self(t) - x

        if not self.final_value < x < self.initial_value:
            raise ValueError(
                f"Can't solve for Pnot = {x}. It is out the calculated range: {self.initial_value}, {self.final_value}"
            )

        result = root_scalar(f, bracket=[self.min_time, self.max_time])
        return result.root

    @lru_cache()
    def _interpolate(self):
        return interp1d(self.times, self.values, bounds_error=False)

    def __call__(self, t: float) -> float:
        return self._interpolate()(t)

    @property
    def acceptor_occ_prob(self) -> Callable:
        return self.hopping_matrix.occ_probs[self.acceptor]

    @staticmethod
    def gen(
        delta_t: float, hopping_matrix: HoppingMatrix, acceptor: int
    ) -> Generator[Tuple, None, None]:
        times = (delta_t * i for i in count())
        matrices = (hopping_matrix(t, delta_t) for t in times)
        A_tilde = (np.delete(np.delete(a, acceptor, axis=0), acceptor, axis=1) for a in matrices)
        p0_tilde = np.delete(hopping_matrix.occ_probs.initial, acceptor)
        #  Taking [a, nb, c, d, ...] -> [(b @ a), (c @ b @ a), (d @ c @ b @ a), ...]
        a_prod = accumulate(A_tilde, lambda a, b: b @ a, initial=p0_tilde)
        pnot = map(np.sum, a_prod)
        return ((delta_t * i, p) for i, p in enumerate(pnot))

    @staticmethod
    def gen_until_time(
        t: float, delta_t: float, hopping_matrix: HoppingMatrix, acceptor: int
    ) -> Iterator[Tuple[float, float]]:

        gen = Pnot.gen(delta_t=delta_t, hopping_matrix=hopping_matrix, acceptor=acceptor)
        while_lt = takewhile(lambda x: x[0] < t, gen)
        extra_terms = islice(gen, 100)
        return chain(while_lt, extra_terms)

    @staticmethod
    def gen_until_prob(
        p: float, delta_t: float, hopping_matrix: HoppingMatrix, acceptor: int
    ) -> Iterator[Tuple[float, float]]:
        gen = Pnot.gen(delta_t=delta_t, hopping_matrix=hopping_matrix, acceptor=acceptor)
        while_gt = takewhile(lambda x: x[1] > p, gen)
        extra_terms = islice(gen, 100)
        return chain(while_gt, extra_terms)

    @staticmethod
    def converged_with_timestep(
        a: HoppingMatrix,
        acceptor: int,
        until_equals: float,
        tolerance: float,
        max_dt: float = 1.0,
        min_dt: float = 1e-4,
    ) -> "Pnot":
        if not max_dt > min_dt:
            raise ValueError(f"Min dt must be below max dt min_dt={min_dt}, max_dt={max_dt}")

        dts = [max_dt]
        last_dt = dts[-1]
        while last_dt > min_dt:
            dts.append(last_dt / 1.61)
            last_dt = dts[-1]

        last_t = -(10.0 ** 6)

        for dt in dts:

            times, p_not_list = zip(
                *Pnot.gen_until_prob(until_equals, acceptor=acceptor, delta_t=dt, hopping_matrix=a)
            )

            p_not = Pnot(
                acceptor=acceptor, times=times, values=p_not_list, delta_t=dt, hopping_matrix=a
            )

            t = p_not.time_when_equal_to(until_equals)
            logger.info(f"For delta_t = {dt:.5f}, Tau = {t:.5f}")
            if isclose(t, last_t, abs_tol=tolerance):
                break
            last_t = t
        else:
            raise ValueError(
                f"Time for Pnot to reach {until_equals} did not converge with timestep to tolerance={tolerance}"
            )
        logger.info(
            f"Yes!! Time for pnot to reach ({until_equals:6f}) successfully converged to {t:6f}"
            f" with difference of {t - last_t:3f}"
            f", which is within the given tolerance, {tolerance}"
        )
        return p_not



import json
from pathlib import Path
from shutil import rmtree
from typing import Union

import attr
import cattr

default_conf_path: Path = Path.home() / ".transit_chem/conf.yaml"


def to_bool(x: Union[str, bool]) -> bool:
    if isinstance(x, str):
        return x.lower() in ("true", "yes")
    elif isinstance(x, bool):
        return x
    else:
        raise TypeError(f"Variable {x} of type {type(x)} can't be converted to bool")


@attr.s()
class Config:
    harmonic_oscillator_max_n: int = attr.ib(default=50, converter=int)
    small_number: float = attr.ib(default=1e-8, converter=float)
    large_number: float = attr.ib(default=1000.0, converter=float)
    float_tol: float = attr.ib(default=1e-6, converter=float)
    enable_progressbar: bool = attr.ib(default=False, converter=to_bool)
    cache_dir: Path = attr.ib(default=Path.home() / ".transit_chem/cache", converter=Path)
    joblib_verbosity: int = attr.ib(default=0, converter=int)


cattr.register_structure_hook(Path, lambda s, t: Path(s))
cattr.register_unstructure_hook(Path, lambda p: str(p))
default_conf = Config()


def load(p: Path = default_conf_path) -> Config:

    conf_file_d = json.loads(p.read_text())
    return attr.evolve(default_conf, **conf_file_d)


try:
    conf: Config = load()
except FileNotFoundError:
    conf: Config = Config()


def set_items(path: Path = default_conf_path, **kwargs):
    new_conf = attr.evolve(conf, **kwargs)
    set_conf(new_conf, path=path)


def show(c: Config = conf):
    name_width = 30
    value_width = 40
    type_width = 10

    print(
        "Name".ljust(name_width)
        + "Value".ljust(value_width)
        + "Type".ljust(type_width)
        + "Default".ljust(value_width)
    )
    total_width = name_width + value_width * 2 + type_width
    print("-" * total_width)
    for field in attr.fields(Config):
        print(
            field.name.ljust(name_width)
            + str(getattr(c, field.name)).ljust(value_width)
            + field.type.__name__.ljust(type_width)
            + str(field.default).ljust(value_width)
        )


def restore_defaults(path: Path = default_conf_path):
    set_conf(c=default_conf, path=path)
    show(default_conf)


def set_conf(c: Config, path: Path = default_conf_path):
    d = cattr.unstructure(c)
    path.write_text(json.dumps(d))
    show(c)
    print(f"Updating config in {path}")


def clear_cache(c: Config = conf):
    rmtree(c.cache_dir)
    print(f"Cleared {c.cache_dir}")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable

import attr
import numpy as np
from scipy import integrate

from transit_chem.molecule import Molecule
from transit_chem.utils import pairwise_array_from_func


# TODO: get the type hints to work for Operator throughout codebase
class Operator(ABC):
    def __call__(self, first: Callable[[float], float], second: Callable[[float], float]) -> float:
        pass

    @property
    @abstractmethod
    def hermitian(self):
        return True

    def matrix(self, basis) -> np.array:
        return pairwise_array_from_func(basis, self, symmetric=self.hermitian)


@attr.s(frozen=True)
class Overlap(Operator):
    lower_limit: float = attr.ib(default=-np.inf)
    upper_limit: float = attr.ib(default=np.inf)
    hermitian = True
    """Compute the overlap integral in 1 dimension over the specified range

    Parameters
    -----------
    first
        The first function
    second
        The second function
    lower_limit
        The lower limit of integration
    upper_limit
        The upper limit of integration

    Returns
    --------
    overlap
        The value of the overlap integral.

    Examples
    ---------

    >>> from transit_chem import overlap
    >>> from numpy import sin, pi
    >>>
    >>> def f1(x):
    ...     return sin(x)
    ...
    >>> def f2(x):
    ...     return sin(2*x)
    ...
    >>> overlap(f1, f2, lower_limit=0, upper_limit=2*pi)
    6.869054119163646e-17
    >>> overlap(f1, f1, lower_limit=0, upper_limit=2*pi)
    3.141592653589793

    """

    def __call__(self, first: Callable, second: Callable) -> float:

        try:
            val = first.__overlap__(second, self.lower_limit, self.upper_limit)
            return val
        except (AttributeError, NotImplementedError):
            pass

        def integrand(x):
            return first(x) * second(x)

        return integrate.quad(integrand, a=self.lower_limit, b=self.upper_limit)[0]


@attr.s()
class Kinetic(Operator):
    overlap: Overlap = attr.ib(default=Overlap())
    hermitian = True

    def __call__(self, first, second) -> float:
        return Overlap()(first, second.__kinetic__())


@attr.s()
class Hamiltonian(Operator):
    potential: Callable[[float], float] = attr.ib()
    kinetic: Kinetic = attr.ib(default=Kinetic())
    hermitian = True

    def __attrs_post_init__(self):
        self.Potential = Potential(self.potential)

    def __call__(self, first, second) -> float:
        return self.Potential(first, second) + self.kinetic(first, second)


@attr.s(frozen=True)
class Potential(Operator):
    potential: Callable[[float], float] = attr.ib()
    overlap: Overlap = attr.ib(default=Overlap())
    hermitian = True

    def __call__(self, first, second) -> float:
        def v(x):
            return self.potential(x) * second(x)

        return self.overlap(first, v)


@attr.s(frozen=True)
class ExtendedHuckelHamiltonian(Operator):
    S: np.array = attr.ib()
    molecule: Molecule = attr.ib()
    hermitian = True

    def __attrs_post_init__(self):
        invalid = [a.element for a in self.molecule if a.element.voie is None]
        if invalid:
            raise ValueError(
                f"Could not make EHT Hamiltonian from elements {invalid}"
                f". They are not configured with VOIE."
            )

    def matrix(self, basis=None) -> np.array:
        """Create the Hamiltonian under the Extended Hueckel Approximation."""

        h = np.zeros(self.S.shape)
        n = self.S.shape[0]

        for i in range(n):
            h[i, i] = -self.molecule.atoms[i].element.voie

        for i in range(n):
            for j in range(n):
                if i != j:
                    h[i, j] = 1.75 * (h[i, i] + h[j, j]) * self.S[i, j] / 2.0
        return h



from functools import partial
from typing import Callable, List

import attr
import numpy as np

from transit_chem.atom import Atom
from transit_chem.utils import pairwise_array_from_func


@attr.s(frozen=True, cmp=False)
class Molecule:
    atoms: List[Atom] = attr.ib()

    def __len__(self):
        return len(self.atoms)

    @property
    def coords(self):
        return np.asarray([a.coords for a in self.atoms])

    @property
    def R(self):
        """Calculate the distances for each atom-atom pair."""
        return pairwise_array_from_func(self.atoms, Atom.distance)

    @property
    def mass(self):
        return sum(a.mass for a in self.atoms)

    @property
    def center_of_mass(self):
        """Determine the center of mass of the molecule."""
        mx = sum(a.x * a.mass for a in self.atoms)
        my = sum(a.y * a.mass for a in self.atoms)
        mz = sum(a.z * a.mass for a in self.atoms)
        return mx / self.mass, my / self.mass, mz / self.mass

    def translated(self, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> "Molecule":
        f = partial(Atom.translated, x=x, y=y, z=z)
        return self.map(f)

    @staticmethod
    def from_xyz(xyz: str) -> "Molecule":
        """Convert from xyzfile and basistype into list of atoms.

        :param xyz_file: Name of xyzfile containing atoms and coordinates
        :type xyz_file: str
        :param basis_type: Descriptor for basis function. Currently supports
        only one type across all atom types.
        :type basis_type: str
        """
        text = (line.split() for line in xyz.splitlines()[2:] if line.strip())

        atoms = [Atom(element=e, x=x, y=y, z=z) for e, x, y, z in text]
        return Molecule(atoms)

    def com_as_origin(self) -> "Molecule":
        x, y, z = self.center_of_mass
        return self.translated(x=-x, y=-y, z=-z)

    def with_atom_aligned_to(self, atom: Atom, x: float, y: float, z: float) -> "Molecule":
        r = atom.rotation_matrix_to(x=x, y=y, z=z)
        return self.rotated(r)

    def rotated(self, r: np.array) -> "Molecule":
        if not r.shape == (3, 3):
            raise ValueError(f"Rotation matrix has invalid shape: {r.shape}")
        rotate = partial(Atom.rotated, r=r)
        return self.map(rotate)

    def map(self, f: Callable) -> "Molecule":
        return attr.evolve(self, atoms=[f(a) for a in self.atoms])

    def scaled(self, factor: float) -> "Molecule":
        return self.map(partial(Atom.scaled, factor=factor))

    def flipped_x(self):
        return self.map(Atom.flipped_x)

    def sorted(self, atomic_key: Callable) -> "Molecule":
        return attr.evolve(self, atoms=sorted(self.atoms, key=atomic_key))

    def rotated_about_x(self, angle: float) -> "Molecule":
        f = partial(Atom.rotated_about_x, angle=angle)
        return self.map(f)

    def rotated_about_y(self, angle: float) -> "Molecule":
        f = partial(Atom.rotated_about_y, angle=angle)
        return self.map(f)

    def rotated_about_z(self, angle: float) -> "Molecule":
        f = partial(Atom.rotated_about_z, angle=angle)
        return self.map(f)

    def __iter__(self):
        return iter(self.atoms)

    def __eq__(self, other):
        return all(a1 == a2 for a1, a2 in zip(self, other))



from scipy.constants import physical_constants as c

HARTREE_TO_EV = c["Hartree energy in eV"][0]
EV_TO_HARTREE = 1.0 / c["Hartree energy in eV"][0]
BOHR_TO_METER = c["Bohr radius"][0]
BOHR_TO_ANGSTROM = 0.529177
ANGSTROM_TO_BOHR = 1.88973
ATU_TO_FEMTOSECOND = c["atomic unit of time"][0] / 1e-15



from typing import Tuple, Union

import attr
import numpy as np

from transit_chem.config import conf
from transit_chem.utils import Parabola
from transit_chem.validation import Range


@attr.s(frozen=True)
class TripleWell:
    center1: Tuple[float, float] = attr.ib()
    barrier12: Tuple[float, float] = attr.ib()
    center2: Tuple[float, float] = attr.ib()
    barrier23: Tuple[float, float] = attr.ib()
    center3: Tuple[float, float] = attr.ib()
    well1: Parabola = attr.ib()
    well2: Parabola = attr.ib()
    well3: Parabola = attr.ib()
    """Construct a well potential from the points.

    x                                                         x
    x                                                         x
    x                                                         x
    x                                                        x
    x                                                        x
     x          barrier12                                   x
     x         x         x            barrier23            x
      x       x          x           x         x          x
       center1            x         x           x       x
                           x       x             center3
                            center2
    """

    def __attrs_post_init__(self):

        # Wells are in expected left to right ordering.
        if (
            not self.center1[0]
            < self.barrier12[0]
            < self.center2[0]
            < self.barrier23[0]
            < self.center3[0]
        ):
            raise ValueError("Points are not in ascending x-value order.")

        # Wells are below the barriers
        assert self.center1[1] < self.barrier12[1]
        assert self.center2[1] < self.barrier12[1]
        assert self.center2[1] < self.barrier23[1]
        assert self.center3[1] < self.barrier23[1]

    def _call_numpy(self, x: np.ndarray) -> np.ndarray:
        y = np.zeros_like(x)

        mask1 = np.where(x < self.barrier12[0])
        y[mask1] = self.well1(x[mask1])

        mask2 = np.where((self.barrier12[0] <= x) & (x <= self.barrier23[0]))
        y[mask2] = self.well2(x[mask2])

        mask3 = np.where(x > self.barrier23[0])
        y[mask3] = self.well3(x[mask3])

        return y

    def _call_scalar(self, x: float) -> float:
        if x < self.barrier12[0]:
            return self.well1(x)
        elif self.barrier12[0] <= x <= self.barrier23[0]:
            return self.well2(x)
        elif x > self.barrier23[0]:
            return self.well3(x)
        else:
            raise ValueError(
                f"Value {x} is not valid for {self}. "
                "Probably the well/barrier parameters are invalid"
            )

    def __call__(self, x: Union[np.ndarray, float]) -> Union[np.ndarray, float]:
        if isinstance(x, np.ndarray):
            return self._call_numpy(x)
        else:
            return self._call_scalar(x)

    @staticmethod
    def from_params(
        well1_depth: float,
        well1_halfwidth: float,
        bridge_length: float,
        bridge_depth: float,
        well3_halfwidth: float,
        well3_depth: float,
    ):
        center1 = (0, 0)
        barrier1 = (center1[0] + well1_halfwidth, center1[1] + well1_depth)
        center2 = (barrier1[0] + 0.5 * bridge_length, barrier1[1] - bridge_depth)
        barrier2 = (barrier1[0] + bridge_length, barrier1[1])
        center3 = (barrier2[0] + well3_halfwidth, barrier2[1] - well3_depth)

        def fit_well(center, barrier):
            center_x, center_y = center
            barrier_x, barrier_y = barrier
            # Third point is reflecting barrier about center
            x = -barrier_x + 2 * center_x
            y = barrier_y
            return Parabola.from_points(center, barrier, (x, y))

        well1 = fit_well(center1, barrier1)
        well2 = Parabola.from_points(barrier1, center2, barrier2)
        well3 = fit_well(center3, barrier2)

        return TripleWell(center1, barrier1, center2, barrier2, center3, well1, well2, well3)


@attr.s(frozen=True)
class Harmonic:
    center: float = attr.ib(validator=[Range(-conf.large_number, conf.large_number)])
    mass: float = attr.ib(default=1.0, validator=Range(conf.small_number, conf.large_number))
    omega: float = attr.ib(default=1.0, validator=Range(conf.small_number, conf.large_number))

    def __call__(self, x: float) -> float:
        return 0.5 * self.mass * (self.omega ** 2) * ((x - self.center) ** 2)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['transit_chem']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=19.3,<20.0',
 'cattrs>=1.0,<2.0',
 'click>=7.0,<8.0',
 'joblib>=0.14.1,<0.15.0',
 'loguru>=0.4.0,<0.5.0',
 'matplotlib>=3.1,<4.0',
 'numba>=0.43.0,<0.44.0',
 'numpy>=1.17,<2.0',
 'scipy>=1.3,<2.0',
 'tqdm>=4.39,<5.0']

entry_points = \
{'console_scripts': ['transit-chem = transit_chem.cli:cli']}

setup_kwargs = {
    'name': 'transit-chem',
    'version': '0.6.1',
    'description': 'Quantifying Probabilistic Electron Transit times.',
    'long_description': 'transit-chem\n============\n\n\n.. image:: https://img.shields.io/pypi/v/transit_chem.svg\n        :target: https://pypi.python.org/pypi/transit_chem\n\n\n.. image:: https://readthedocs.org/projects/transit-chem/badge/?version=latest\n        :target: https://transit-chem.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\nQuantifying electron transit in donor-bridge-acceptor systems using probabilistic confidence.\n\n\n* Free software: MIT license\n* Documentation: https://transit-chem.readthedocs.io.\n\n\nFeatures\n--------\n',
    'author': 'Evan Curtin',
    'author_email': 'fakeemail@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



# import numpy as np
#
# import transittime
#
#
# def main(argv=None):
#
#     x = np.linspace(-10, 10, 500)
#     one = transittime.OneD(coordinates=x, nregions=3)
#     # t = np.linspace(0, 1500, 1500)
#     initial_state = one.basis_funcs[0](one.coords, 0, 0.0)
#     one.calc_all(initialstate=initial_state)
#
#     # fname = '../analysis/xyz_files/2C.xyz'
#     # mol = transittime.Molecule(fname)
#     # mol.trim_hydrogens()
#     # mol.scale(transittime.units.conversion_factors['ANGSTROM_TO_BOHR'])
#
#     #  Aligns the line joining the two nitrogens with the x axis.
#     # nitrogens = [atom for atom in mol if atom.symbol == 'N']
#     # if nitrogens:
#     #    mol.align_to_x(*nitrogens)
#     # else:
#     #    raise AttributeError('Did not find nitrogens!')
#
#     #  Align the nitro (acceptor) group to be +x
#     # oxygens = [atom for atom in mol if atom.symbol == 'O']
#     # if oxygens:
#     #    if all(oxygen.coords[0] < 0 for oxygen in oxygens):
#     #        mol.flip_x()
#
#     #  Sort into ascending order of x coordinates for convenience
#     # mol.sort(atomic_key=lambda atom: atom.x)
#
#     #  This works since the molecule is aligned/shifted as expected.
#     # donor_n = min(nitrogens, key=lambda atom: atom.x)
#     # acceptor_n = max(nitrogens, key=lambda atom: atom.x)
#     # donor_index = mol.atoms.index(donor_n)
#     # acceptor_index = mol.atoms.index(acceptor_n)
#
#     #  Since they are aligned on the x axis already and atoms are sorted by x value, the adjacent atoms in
#     #  the list will be the relevant carbons for the boundary
#
#     # donor_c = mol.atoms[donor_index + 1]
#     # acceptor_c = mol.atoms[acceptor_index - 1]
#
#     # donor_bridge_bound = 0.5 * (donor_n.x + donor_c.x)
#     # acceptor_bridge_bound = 0.5 * (acceptor_n.x + acceptor_c.x)
#     # bounds = (donor_bridge_bound, acceptor_bridge_bound)
#
#     # ThreeDCalc = transittime.threeD.ThreeD(nregions=3, boundaries=bounds, molecule=mol, id='testing')
#     # ThreeDCalc.calc_time_independent()
#
#     #  Initial state is Pz orbital on donor nitrogen. This was most convenient way to do it.
#     # initialstate = transittime.basis.PzBasis(*donor_n.coords, donor_n.alpha)
#     # ThreeDCalc.set_initial_state(initialstate)  # does overlap integrals, can be slow
#     # ThreeDCalc.calc_time_dependent(maxsteps=10**6)
#
#
# if __name__ == "__main__":
#     import sys
#
#     main(sys.argv)



from math import isinf, isnan
from typing import Any, Callable

import attr
from attr import Attribute

Validator = Callable[[Any, Attribute], Any]


__all__ = ["Range", "not_inf", "not_nan"]


class ValidationError(ValueError):
    pass


@attr.s
class Range:
    """Attrs Validator to ensure values between a min and a max value.

    Parameters
    -----------
    min
        The minimum allowed value
    max
        The maximum allowed value

    """

    min = attr.ib()
    max = attr.ib()

    @min.validator
    def check(self, attribute, value):
        if not self.min <= self.max:
            raise ValidationError(f"Min must be below max got min={self.min}, max={self.max}")

    def __call__(self, instance: Any, attribute: Attribute, value: Any) -> Any:
        if not self.min <= value <= self.max:
            raise ValidationError(
                f"Validation error for {instance}. "
                f"{attribute.name} out of range, must be "
                f"between {self.min} and {self.max}, got {value}"
            )


def not_nan(instance: Any, attribute: Attribute, value: float):
    if isnan(value):
        raise ValidationError(
            f"Validation error for {instance}. {attribute.name} is NaN but that's not allowed"
        )


def not_inf(instance: Any, attribute: Attribute, value: float):
    if isinf(value):
        raise ValidationError(
            f"Validation error for {instance}. {attribute.name} is Infinite but that's not allowed"
        )
